#!/usr/bin/perl

##++
##     CGI Lite v1.9
##     Last modified: Apr 17, 2000 (BDL - see HISTORY)
##
##     Copyright (c) 1995, 1996, 1997 by Shishir Gundavaram
##     All Rights Reserved
##
##     E-Mail: shishir@ora.com
##
##     Permission  to  use,  copy, and distribute is hereby granted,
##     providing that the above copyright notice and this permission
##     appear in all copies and in supporting documentation.
##--

package CGI_Lite;
require 5.002;
require Exporter;

@ISA    =    (Exporter);
@EXPORT = qw (browser_escape
              url_encode
              url_decode
              is_dangerous
              escape_dangerous_chars);

##++
## Global Variables
##--

$CGI_Lite::VERSION = '1.9'; # BDL

##++
##  Start
##--

sub new
{
    my $self;

    $self = {
	        multipart_dir    =>    undef,
	        default_dir      =>    '/tmp',
	        file_type        =>    'name',
	        platform         =>    'Windows',
	        buffer_size      =>    1024,
	        timestamp        =>    1,
		filter           =>    undef,
	        web_data         =>    {},
		ordered_keys     =>    [],
		all_handles      =>    [],
	        error_status     =>    0,
	        error_message    =>    undef,
		file_size_limit	 =>    2097152,
	    };

    $self->{convert} = { 
	                   'text/html'    => 1,
	                   'text/plain'   => 1
	               };

    $self->{file} = { Unix => '/',    Mac => ':',    PC => '\\'       };
    $self->{eol}  = { Unix => "\012", Mac => "\015", PC => "\015\012" };

    bless $self;
    return $self;
}

sub Version 
{ 
    return $VERSION;
}

sub set_directory
{
    my ($self, $directory) = @_;

    stat ($directory);

    if ( (-d _) && (-e _) && (-r _) && (-w _) ) {
	$self->{multipart_dir} = $directory;
	return (1);

    } else {
	return (0);
    }
}

sub add_mime_type
{
    my ($self, $mime_type) = @_;

    $self->{convert}->{$mime_type} = 1 if ($mime_type);
}

sub remove_mime_type
{
    my ($self, $mime_type) = @_;

    if ($self->{convert}->{$mime_type}) {
	delete $self->{convert}->{$mime_type};
	return (1);

    } else {
	return (0);
    }
}

sub get_mime_types
{
    my $self = shift;

    return (sort keys %{ $self->{convert} });
}

sub set_platform
{
    my ($self, $platform) = @_;

    if ($platform =~ /(?:PC|NT|Windows(?:95)?|DOS)/i) {
        $self->{platform} = 'PC';

    } elsif ($platform =~ /Mac(?:intosh)?/i) {

	## Should I check for NeXT here :-)

        $self->{platform} = 'Mac';

    } else {
	$self->{platform} = 'Unix';
    }
}

sub set_file_type
{
    my ($self, $type) = @_;

    if ($type =~ /^handle$/i) {
	$self->{file_type} = 'handle';
    } else {
	$self->{file_type} = 'name';
    }
}

sub add_timestamp
{
    my ($self, $value) = @_;

    if ( ($value < 0) || ($value > 2) ) {
	$self->{timestamp} = 1;
    } else {
	$self->{timestamp} = $value;
    }
}

sub filter_filename
{
    my ($self, $subroutine) = @_;

    $self->{filter} = $subroutine;
}

sub set_buffer_size
{
    my ($self, $buffer_size) = @_;
    my $content_length;

    $content_length = $ENV{CONTENT_LENGTH} || return (0);

    if ($buffer_size < 256) {
	$self->{buffer_size} = 256;
    } elsif ($buffer_size > $content_length) {
	$self->{buffer_size} = $content_length;
    } else {
	$self->{buffer_size} = $buffer_size;
    }

    return ($self->{buffer_size});
}

sub parse_new_form_data
# Reset state before parsing (for persistant CGI objects, e.g. under FastCGI) 
# BDL
{
	my ($self, @param) = @_;

	# close files (should happen anyway when 'all_handles' is cleared...)
	$self->close_all_files();

	$self->{web_data}	= {};
	$self->{ordered_keys} 	= [];
	$self->{all_handles} 	= [];
	$self->{error_status} 	= 0;
	$self->{error_message} 	= undef;

	$self->parse_form_data(@param);
}

sub parse_form_data
{
    my ($self, $user_request) = @_;
    my ($request_method, $content_length, $content_type, $query_string,
	$boundary, $post_data, @query_input);

    $request_method = $user_request || $ENV{REQUEST_METHOD} || '';
    $content_length = $ENV{CONTENT_LENGTH};
    $content_type   = $ENV{CONTENT_TYPE};

    if ($request_method =~ /^(get|head)$/i) {

	$query_string = $ENV{QUERY_STRING};
	$self->_decode_url_encoded_data (\$query_string, 'form');

	return wantarray ?
	    %{ $self->{web_data} } : $self->{web_data};

    } elsif ($request_method =~ /^post$/i) {

	if (!$content_type || 
	    ($content_type eq 'application/x-www-form-urlencoded')) {

	    local $^W = 0;

	    read (STDIN, $post_data, $content_length);
	    $self->_decode_url_encoded_data (\$post_data, 'form');

	    return wantarray ? 
		%{ $self->{web_data} } : $self->{web_data};

	} elsif ($content_type =~ /multipart\/form-data/) {
	    ($boundary) = $content_type =~ /boundary=(\S+)$/;
	    $self->_parse_multipart_data ($content_length, $boundary);

	    return wantarray ? 
		%{ $self->{web_data} } : $self->{web_data};

	} else {
	    $self->_error ('Invalid content type!');
	}

    } else {

	##++
	##  Got the idea of interactive debugging from CGI.pm, though it's
        ##  handled a bit differently here. Thanks Lincoln!
	##--

	print "[ Reading query from standard input. Press ^D to stop! ]\n";

	@query_input = <>;
	chomp (@query_input);

	$query_string = join ('&', @query_input);
	$query_string =~ s/\\(.)/sprintf ('%%%x', ord ($1))/eg;
 
	$self->_decode_url_encoded_data (\$query_string, 'form');

	return wantarray ?
	    %{ $self->{web_data} } : $self->{web_data};
    }
}

sub parse_cookies
{
    my $self = shift;
    my $cookies;

    $cookies = $ENV{HTTP_COOKIE} || return;

    $self->_decode_url_encoded_data (\$cookies, 'cookies');

    return wantarray ? 
        %{ $self->{web_data} } : $self->{web_data};
}

sub get_ordered_keys
{
    my $self = shift;

    return wantarray ?
	@{ $self->{ordered_keys} } : $self->{ordered_keys};
}

sub print_data
{
    my $self = shift;
    my ($key, $value, $eol);

    $eol = $self->{eol}->{$self->{platform}};

    foreach $key (@{ $self->{ordered_keys} }) {
	$value = $self->{web_data}->{$key};

	if (ref $value) {
	    print "$key = @$value$eol";
	} else {
	    print "$key = $value$eol";
	}
    }
}

sub print_mime_type
{
    my ($self, $field) = @_;

    return($self->{'mime_types'}->{$field});
}

*print_form_data = *print_cookie_data = \&print_data;

sub wrap_textarea
{
    my ($self, $string, $length) = @_;
    my ($new_string, $platform, $eol);

    $length     = 70 unless ($length);
    $platform   = $self->{platform};
    $eol        = $self->{eol}->{$platform};
    $new_string = $string || return;
	
    $new_string =~ s/[\0\r]\n?/ /sg;
    $new_string =~ s/(.{0,$length})\s/$1$eol/sg;

    return $new_string;
}

sub get_multiple_values
{
    my ($self, $array) = @_;

    return (ref $array) ? (@$array) : $array;
}

sub create_variables
{
    my ($self, $hash) = @_;
    my ($package, $key, $value);
    
    $package = $self->_determine_package;

    while (($key, $value) = each %$hash) {
	${"$package\:\:$key"} = $value;
    }
}

sub is_error
{
    my $self = shift;

    if ($self->{error_status}) {
	return (1);
    } else {
	return (0);
    }
}

sub get_error_message
{
    my $self = shift;

    return $self->{error_message} if ($self->{error_message});
}

sub return_error
{
    my ($self, @messages) = @_;

    print "@messages\n";

    exit (1);
}

##++
##  Exported Subroutines
##--

sub browser_escape
{
    my $string = shift;

    $string =~ s/([<&"#%>])/sprintf ('&#%d;', ord ($1))/ge;

    return $string;
}

sub url_encode
{
    my $string = shift;

    $string =~ s/([\x00-\x20"#%;<>?{}|\\\\^~`\[\]\x7F-\xFF])/
                 sprintf ('%%%x', ord ($1))/eg;

    return $string;
}

##+
##  Thanks to Paul Phillips for the meta-character list.
##--

sub url_decode
{
    my $string = shift;

    $string =~ s/%([\da-fA-F]{2})/chr (hex ($1))/eg;

    return $string;
}

sub is_dangerous
{
    my $string = shift;

    if ($string =~ /[;<>\*\|`&\$!#\(\)\[\]\{\}:'"]/) {
        return (1);
    } else {
        return (0);
    }
}

sub escape_dangerous_chars
{
    my $string = shift;

    $string =~ s/([;<>\*\|`&\$!#\(\)\[\]\{\}:'"])/\\$1/g;

    return $string;
}

##++
##  Internal Methods
##--

sub _error
{
    my ($self, $message) = @_;

    $self->{error_status}  = 1;
    $self->{error_message} = $message;
}

sub _determine_package
{
    my $self = shift;
    my ($frame, $this_package, $find_package);

    $frame = -1;
    ($this_package) = split (/=/, $self);

    do {
	$find_package = caller (++$frame);
    } until ($find_package !~ /^$this_package/);

    return ($find_package);
}   

##++
##  Decode URL encoded data
##--

sub _decode_url_encoded_data
{
    my ($self, $reference_data, $type) = @_;
    my $code;

    $code = <<'End_of_URL_Decode';

    my (@key_value_pairs, $delimiter, $key_value, $key, $value);

    @key_value_pairs = ();

    return unless ($$reference_data);

    if ($type eq 'cookies') {
	$delimiter = ';\s+';
    } else {
	$delimiter = '&';
    }

    $$reference_data =~ tr/+/ /;
    @key_value_pairs = split (/$delimiter/, $$reference_data);
		
    foreach $key_value (@key_value_pairs) {
	($key, $value) = split (/=/, $key_value, 2);

	$value = '' unless defined $value;	# avoid 'undef' warnings for "key=" BDL Jan/99

	$key   =~ s/%([\da-fA-F]{2})/chr (hex ($1))/eg;
	$value =~ s/%([\da-fA-F]{2})/chr (hex ($1))/eg;
	
	if ( defined ($self->{web_data}->{$key}) ) {
	    $self->{web_data}->{$key} = [$self->{web_data}->{$key}] 
	        unless ( ref $self->{web_data}->{$key} );

	    push (@{ $self->{web_data}->{$key} }, $value);
	} else {
	    $self->{web_data}->{$key} = $value;
	    push (@{ $self->{ordered_keys} }, $key);
	}
    }

End_of_URL_Decode

    eval ($code);
    $self->_error ($@) if $@;
}

##++
##  Methods dealing with multipart data
##--

sub _parse_multipart_data
{
    my ($self, $total_bytes, $boundary) = @_;
    my ($code, $files);

    local $^W = 0;
    $files    = {};

    $code = <<'End_of_Multipart';

    my ($seen, $buffer_size, $byte_count, $platform, $eol, $handle, 
	$directory, $bytes_left, $buffer_size, $new_data, $old_data, 
	$current_buffer, $changed, $store, $disposition, $headers, 
        $mime_type, $convert, $field, $file, $new_name, $full_path);

    $seen        = {};
    $buffer_size = $self->{buffer_size};
    $byte_count  = 0;
    $platform    = $self->{platform};
    $eol         = $self->{eol}->{$platform};
    $handle      = 'CL00';
    $directory   = $self->{multipart_dir} || $self->{default_dir};

    binmode STDIN; # added for binaries

    while (1) {
	if ( ($byte_count < $total_bytes) &&
	     (length ($current_buffer) < ($buffer_size * 2)) ) {

	    $bytes_left  = $total_bytes - $byte_count;
	    $buffer_size = $bytes_left if ($bytes_left < $buffer_size);

	    read (STDIN, $new_data, $buffer_size);
            $self->_error ("Oh, Oh! I'm upset! Can't read what I want.")
		if (length ($new_data) != $buffer_size);

	    $byte_count += $buffer_size;

	    if ($old_data) {
		$current_buffer = join ('', $old_data, $new_data);
	    } else {
		$current_buffer = $new_data;
	    }

	} elsif ($old_data) {
	    $current_buffer = $old_data;
	    $old_data = undef;

	} else {
	    last;
	}

	$changed = 0;

	##++
	##  When Netscape Navigator creates a random boundary string, you
	##  would expect it to pass that _same_ value in the environment
	##  variable CONTENT_TYPE, but it does not! Instead, it passes a
	##  value that has the first two characters ("--") missing.
	##--

	if ($current_buffer =~ 
            /(.*?)(?:\015?\012)?-*$boundary-*[\015\012]*(?=(.*))/os) {

	    ($store, $old_data) = ($1, $2);

            if ($current_buffer =~ 
             /[Cc]ontent-[Dd]isposition: ([^\015\012]+)\015?\012  # Disposition
              (?:([A-Za-z].*?)(?:\015?\012){2})?                  # Headers
              (?:\015?\012)?                                      # End
              (?=(.*))                                            # Other Data
             /xs) {

		($disposition, $headers, $current_buffer) = ($1, $2, $3);
		$old_data = $current_buffer;

		($mime_type) = $headers =~ /[Cc]ontent-[Tt]ype: (\S+)/;

		$self->_store ($platform, $file, $convert, $handle, $eol, 
			       $field, \$store, $seen);

		close ($handle) if (fileno ($handle));

		if ($mime_type && $self->{convert}->{$mime_type}) {
		    $convert = 1;
		} else {
		    $convert = 0;
		}

		$changed = 1;

		($field) = $disposition =~ /name="([^"]+)"/;
		++$seen->{$field};

		$self->{'mime_types'}->{$field} = $mime_type;

                if ($seen->{$field} > 1) {
                    $self->{web_data}->{$field} = [$self->{web_data}->{$field}]
                        unless (ref $self->{web_data}->{$field});
                } else {
                    push (@{ $self->{ordered_keys} }, $field);
                }

                if (($file) = $disposition =~ /filename="(.*)"/) {
                    $file =~ s|.*[:/\\](.*)|$1|;

                    $new_name = $self->_get_file_name ($platform,
                                                       $directory, $file);

                    $self->{web_data}->{$field} = $new_name;

                    $full_path = join ($self->{file}->{$platform}, 
                                       $directory, $new_name);

                    unlink($full_path);		# added 2009/06/01
                    open (++$handle, ">$full_path") 
	                || $self->_error ("Can't create file: $full_path!");

                    binmode $handle; 		# added for binaries
                    $files->{$new_name} = $full_path;
                } 
            }

	} elsif ($old_data) {
            $store    = $old_data;
            $old_data = $new_data;

	} else {
	    $store          = $current_buffer;
            $current_buffer = $new_data;
        }

        unless ($changed) {
           $self->_store ($platform, $file, $convert, $handle, $eol, 
                          $field, \$store, $seen);
        }
    }

    close ($handle) if (fileno ($handle));

End_of_Multipart

    eval ($code);
    $self->_error ($@) if $@;

    $self->_create_handles ($files) if ($self->{file_type} eq 'handle');
}

sub _store
{
    my ($self, $platform, $file, $convert, $handle, $eol, $field, 
	$info, $seen) = @_;

    if ($file) {
	if ($convert) {
	    $$info =~ s/\015\012/$eol/og;
#	    $$info =~ s/\015\012/$eol/og  if ($platform ne 'PC');
#	    $$info =~ s/\015/$eol/og      if ($platform ne 'Mac');
#	    $$info =~ s/\012/$eol/og      if ($platform ne 'Unix');
	}

    	print $handle $$info;

    } elsif ($field) {
	if ($seen->{$field} > 1) {
	    $self->{web_data}->{$field}->[$seen->{$field}-1] .= $$info;
	} else {
	    $self->{web_data}->{$field} .= $$info;
        }
    }
}

sub _get_file_name
{
    my ($self, $platform, $directory, $file) = @_;
    my ($filtered_name, $filename, $timestamp, $path);

    $filtered_name = &{ $self->{filter} }($file)
        if (ref ($self->{filter}) eq 'CODE');

    $filename  = $filtered_name || $file;
    $timestamp = time . '__' . $filename;

    if (!$self->{timestamp}) {
	return $filename;

    } elsif ($self->{timestamp} == 1) {
	return $timestamp;
	
    } elsif ($self->{timestamp} == 2) {
	$path = join ($self->{file}->{$platform}, $directory, $filename);
	
	return (-e $path) ? $timestamp : $filename;
    }
}

sub _create_handles
{
    my ($self, $files) = @_;
    my ($package, $handle, $name, $path);

    $package = $self->_determine_package;

    while (($name, $path) = each %$files) {
	$handle = "$package\:\:$name";
	open ($handle, "<$path")
            || $self->_error ("Can't read file: $path!");

	binmode $handle;	# added for binaries;
	push (@{ $self->{all_handles} }, $handle);
    }
}

sub close_all_files
{
    my $self = shift;
    my $handle;

    foreach $handle (@{ $self->{all_handles} }) {
	close $handle;
    }
}

1;

